from enum import Enum, auto
from pydantic.fields import FieldInfo
from pydantic._internal._fields import PydanticGeneralMetadata
from typing import Any, Callable, Dict, Optional, TypeVar, Union, get_args, get_origin

T = TypeVar("T")


class Grammar(Enum):
    EBNF = auto()
    GBNF = auto()

    def rule(self, name: str, definition: str) -> str:
        if self == Grammar.EBNF:
            return f"{name} = {definition}"
        elif self == Grammar.GBNF:
            return f"{name} ::= {definition}"
        assert False

    def from_types(self, types_: list[Union[str, type]]) -> str:
        rules: dict[str, str] = {}

        def _rec_type_to_grammar(type_: type) -> str:
            basetype = Type(type_)
            if basetype.name in rules:
                return basetype.name
            rules[basetype.name] = basetype.definition(self)
            for arg in basetype.args:
                _rec_type_to_grammar(arg)
            return basetype.name

        rules["root"] = " ".join([_rec_type_to_grammar(x) if not isinstance(x, str) else f'"{x}"' for x in types_])
        return "\n".join([self.rule(n, r) for n, r in rules.items()])


class TypeHandler:
    _BASE_TYPES: dict[type, "TypeHandler"] = {}
    _COMPOSED_TYPES: list["TypeHandler"] = []

    @classmethod
    def get(cls, type_: type) -> Optional["TypeHandler"]:
        if type_ in cls._BASE_TYPES:
            return cls._BASE_TYPES[type_]
        for t in cls._COMPOSED_TYPES:
            if callable(t._type_match) and t._type_match(type_):
                return t
        return None

    def __init__(
        self,
        type_match: Union[type, Callable[[type], bool]],
        name: Union[str, Callable[[type], str]],
        ebnf: Union[str, Callable[[type], str]],
        gbnf: Union[str, Callable[[type], str]],
        parser: Callable[[str, type[T]], tuple[Optional[T], int]],
        get_args: Optional[Callable[[type], list[type]]] = None,
    ) -> None:
        self._type_match = type_match
        self._name = name
        self._ebnf = ebnf
        self._gbnf = gbnf
        self._parser = parser
        self._get_args = get_args
        if type(type_match) is type:
            self._BASE_TYPES[type_match] = self
        else:
            assert callable(type_match)
            self._COMPOSED_TYPES.append(self)

    def name(self, type_: type) -> str:
        return self._name(type_) if callable(self._name) else self._name

    def ebnf(self, type_: type) -> str:
        return self._ebnf(type_) if callable(self._ebnf) else self._ebnf

    def gbnf(self, type_: type) -> str:
        return self._gbnf(type_) if callable(self._gbnf) else self._gbnf

    def parse(self, payload: str, type_: type[T]) -> tuple[Optional[T], int]:
        return self._parser(payload, type_)

    def args(self, type_: type) -> list[type]:
        return self._get_args(type_) if self._get_args else []


class Type:
    def __init__(self, type_: type) -> None:
        handler = TypeHandler.get(type_)
        if not handler:
            raise Exception('Unsupported type "%r"' % type_)
        self._type = type_
        self._handler = handler
        self._name = self._handler.name(self._type)
        self._args = self._handler.args(self._type)

    @property
    def name(self) -> str:
        return self._name

    @property
    def args(self) -> list[type]:
        return self._args

    def definition(self, grammar: Grammar) -> str:
        if grammar == Grammar.EBNF:
            return self._handler.ebnf(self._type)
        elif grammar == Grammar.GBNF:
            return self._handler.gbnf(self._type)
        assert False

    def parse(self, payload: str) -> tuple[Any, int]:
        obj, i = self._handler.parse(payload, self._type)
        if obj is None:
            raise Exception('Could not parse type "%r" in payload:\n%s' % (self._type, payload))
        return obj, i


# Pydantic annotated types need a proper rule name in our BNF.
# Since such type may contain a lot of information, we prefer to number them (Fieldinfo1,...)
# We keep a dict of previously seen FieldInfo with an associated counter.
# Unfortunately, FieldInfo doesn't hash properly, we have to implement our how hash function.
FIELD_HASHES: Dict[int, int] = {}


def _hash_fieldinfo(field: FieldInfo) -> int:
    hashes = []
    for m in field.metadata:
        if isinstance(m, PydanticGeneralMetadata):
            hashes.append(hash(frozenset([(k, v) for k, v in m.__dict__.items()])))
        else:
            hashes.append(hash(m))
    return hash(frozenset(hashes))


def composed_name(type_: type[T]) -> str:
    orig = get_origin(type_)
    if orig:
        return str(orig.__name__.lower() + "".join([type_name(a).title() for a in get_args(type_)]))
    return type_.__name__.lower()


def type_name(type_: type) -> str:
    if isinstance(type_, FieldInfo):
        h = _hash_fieldinfo(type_)
        if h not in FIELD_HASHES:
            FIELD_HASHES[h] = len(FIELD_HASHES.keys())
        return f"Fieldinfo{FIELD_HASHES[h]}"
    return Type(type_).name



import re
import inspect

from typing import Any, Optional, Union
from llama_cpp.llama_grammar import LlamaGrammar

from llama_prompter.types import Grammar, PrompterVarType, Type

TEMPLATE_VAR_RE = re.compile(r"(?<!{){([a-zA-Z_]+):([^{}]*)}(?!})")


class Variable:
    def __init__(self, name: str, type_: PrompterVarType) -> None:
        self._name = name
        self._type = type_


def get_caller_symbols() -> tuple[dict[str, str], dict[str, str]]:
    frame = inspect.currentframe()
    if frame is None or frame.f_back is None or frame.f_back.f_back is None:
        raise Exception("Could not retrieve caller's symbols")
    return frame.f_back.f_back.f_globals, frame.f_back.f_back.f_locals


class Prompter:
    def __init__(self, template: str) -> None:
        self._prompt = ""
        self._sequences: list[Union[str, Variable]] = []
        caller_g, caller_l = get_caller_symbols()
        matches = [x for x in TEMPLATE_VAR_RE.finditer(template)]
        for i, m in enumerate(matches):
            if i == 0:
                self._prompt = template[0 : m.start(0)]  # noqa: E203
            else:
                self._sequences.append(template[matches[i - 1].end(0) : m.start(0)])  # noqa: E203
            type_ = eval(m[2], caller_g, caller_l)
            assert isinstance(type_, type)
            var = Variable(m[1], type_)
            self._sequences.append(var)
        if matches and matches[-1].end(0) < len(template):
            self._sequences.append(template[matches[-1].end(0) :])  # noqa: E203
        self._grammar = (
            LlamaGrammar.from_string(Grammar.GBNF.from_types([x._type if isinstance(x, Variable) else x for x in self._sequences]), verbose=False)
            if self._sequences
            else None
        )

    @property
    def prompt(self) -> str:
        return self._prompt

    @property
    def grammar(self) -> Optional[LlamaGrammar]:
        return self._grammar

    def decode_response(self, response: str) -> dict[str, Any]:
        variables: dict[str, Any] = {}
        start = 0
        for i in self._sequences:
            if isinstance(i, str):
                if not response[start:].startswith(i):
                    raise Exception("Could not decode")
                start += len(i)
            elif isinstance(i, Variable):
                obj, idx = Type(i._type).parse(response[start:])
                variables[i._name] = obj
                start += idx
        return variables



from typing import Annotated, Any, Union
from llama_prompter.types.base import Grammar, Type

import llama_prompter.types.base_types
import llama_prompter.types.list
import llama_prompter.types.tuple
import llama_prompter.types.dict
import llama_prompter.types.pydantic
import llama_prompter.types.annotated  # noqa: F401

__all__ = [
    "Grammar",
    "Type",
]

PrompterVarType = Union[type, Annotated[Any, ...]]



from typing import Annotated, Optional, TypeVar, get_args, get_origin

from llama_prompter.types.base import TypeHandler, Type, composed_name

T = TypeVar("T")


def _annotated_ebnf(type_: type) -> str:
    orig_type, annotation = get_args(type_)
    if orig_type == str:
        for metadata in annotation.metadata:
            if hasattr(metadata, "pattern"):
                return r"~'\"" + str(metadata.pattern) + "\"'"
    assert False


def _annotated_gbnf(type_: type) -> str:
    orig_type, annotation = get_args(type_)
    if orig_type == str:
        for metadata in annotation.metadata:
            if hasattr(metadata, "pattern"):
                return '"\\"" (' + str(metadata.pattern) + ') "\\""'
    assert False


def _annotated_parser(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
    return Type(get_args(type_)[0]).parse(payload)


TypeHandler(
    lambda x: get_origin(x) == Annotated,
    composed_name,
    _annotated_ebnf,
    _annotated_gbnf,
    _annotated_parser,
)



import json

from typing import Optional, TypeVar, get_args, get_origin

from llama_prompter.types.base import TypeHandler, composed_name, type_name

T = TypeVar("T")


def dict_parser(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
    d, i = json.JSONDecoder().raw_decode(payload)
    return type_(**d), i


TypeHandler(
    lambda x: get_origin(x) == dict,
    composed_name,
    lambda x: (kv := f'{type_name(get_args(x)[0])} ":" {type_name(get_args(x)[1])}', f'"{{" ({kv} ("," {kv})*)? "}}"')[-1],
    lambda x: (kv := f'{type_name(get_args(x)[0])} ":" {type_name(get_args(x)[1])}', f'"{{" ({kv} ("," {kv})*)? "}}"')[-1],
    dict_parser,
    get_args=lambda x: [z for z in get_args(x)],
)



from typing import Any, Optional, TypeVar, get_args, get_origin

from llama_prompter.types.base import Type, TypeHandler, composed_name, type_name

T = TypeVar("T")


def _list_parser(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
    start = 1  # skip first "["
    result: list[Any] = []
    arg_type = get_args(type_)[0]
    while payload[start - 1] != "]":
        if start >= len(payload):
            return None, 0
        arg, i = Type(arg_type).parse(payload[start:])
        if arg is None:
            return None, 0
        result.append(arg)
        start += i + 1  # skip comma ","
    return list(result), start  # type: ignore


TypeHandler(
    lambda x: get_origin(x) == list,
    composed_name,
    lambda x: (arg := type_name(get_args(x)[0]), f'"[" ({arg} ("," {arg})*)? "]"')[-1],
    lambda x: (arg := type_name(get_args(x)[0]), f'"[" ({arg} ("," {arg})*)? "]"')[-1],
    _list_parser,
    get_args=lambda x: [z for z in get_args(x)],
)



from typing import Any, Optional, TypeVar, get_args, get_origin

from llama_prompter.types.base import Type, TypeHandler, composed_name, type_name

T = TypeVar("T")


def _tuple_parser(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
    start = 1  # skip first "["
    result: list[Any] = []
    for arg_type in get_args(type_):
        if start >= len(payload):
            return None, 0
        arg, i = Type(arg_type).parse(payload[start:])
        if arg is None:
            return None, 0
        result.append(arg)
        start += i + 1  # skip comma ","
    return tuple(result), start  # type: ignore


TypeHandler(
    lambda x: get_origin(x) == tuple,
    composed_name,
    lambda x: '"[" ' + ' "," '.join([type_name(a) for a in get_args(x)]) + ' "]"',
    lambda x: '"[" ' + ' "," '.join([type_name(a) for a in get_args(x)]) + ' "]"',
    _tuple_parser,
    get_args=lambda x: [z for z in get_args(x)],
)



import re

from typing import Any, Callable, Optional, TypeVar

from llama_prompter.types.base import TypeHandler

T = TypeVar("T")


def _regex_parser(regex: str, cast: Callable[[str], Any]) -> Callable[[str, type[T]], tuple[Optional[T], int]]:
    r = re.compile(regex)

    def _parser_callback(payload: str, type_: type[T]) -> tuple[Optional[T], int]:
        m = r.match(payload)
        if not m:
            return None, 0
        return cast(m.group(1)), len(m.group(0))

    return _parser_callback


TypeHandler(
    bool,
    "boolean",
    r'"true" / "false"',
    '("true" | "false")',
    _regex_parser(r"(true|false)", lambda x: x == "true"),
)
TypeHandler(
    float,
    "float",
    r"~'-?[0-9]*(.[0-9]*)?'",
    '("-"? ([0-9]*) ("." [0-9]+))',
    _regex_parser(r"(\-?[0-9]*\.[0-9]+)", float),
)
TypeHandler(
    int,
    "integer",
    r"~'-?[0-9]+'",
    '("-"? ([0-9]+))',
    _regex_parser(r"(\-?[0-9]+)", int),
)
TypeHandler(
    str,
    "string",
    r"~'\"[0-9 a-z]*\"'i",
    '"\\"" ([^"\\\\] | "\\\\" (["\\\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]))* "\\""',
    _regex_parser(r'"((?:\\.|[^"\\])*)"', str),
)



from pydantic import BaseModel

from llama_prompter.types.base import TypeHandler, composed_name, type_name
from llama_prompter.types.dict import dict_parser


def _basemodel_bnf(type_: type) -> str:
    assert issubclass(type_, BaseModel)
    return '"{" ' + ' "," '.join([f'"\\"{n}\\":" {type_name(f.annotation)}' for n, f in type_.model_fields.items() if f.annotation]) + ' "}"'


def _basemodel_getargs(type_: type) -> list[type]:
    assert issubclass(type_, BaseModel)
    return [f.annotation for _, f in type_.model_fields.items() if f.annotation]


TypeHandler(
    lambda x: hasattr(x, "__subclasses__") and issubclass(x, BaseModel),
    composed_name,
    _basemodel_bnf,
    _basemodel_bnf,
    dict_parser,
    get_args=_basemodel_getargs,
)
